// Copyright 2004 Laurent GUEGUEN

// This file is part of Sarment.

// Sarment is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// Sarment is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with Sarment; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include "matrice.h"
#include "sequence.h"
 
#include "mat_iter.h"
#include "partition.h"

#include "lexique.h"
#include "RM_lex.h"
#include "posi_motif.h"
#include "math.h"

#include <ctype.h>
#include <stdlib.h>

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Cmatrice::Cmatrice(const Fmatrice& M) : Fmatrice(M)
{
  int i;
  if (_nvct){
    _seq = new vector<Cvaleur>[_nvct];
    for (i=0;i<_nvct;i++)
      _seq[i].construit_a_vide(_nval);
    
    _fin=&_seq[_nvct-1];
    for (i=0;i<_nvct;i++)
      _seq[i].recup_prem(M._seq[i],_nval);
  }
  else
    _seq=_fin=0;
}

void Cmatrice::genere(const int i, const string& s)
{
  nettoie();
  if (i<=0)
    return;

  int l=s.length();
  int j=-1;
  while (++j<l)
    if (s.find(s[j],j+1)!=string::npos)
      return;

  _nval=l;
  _nvct=i;
  _seq= new vector<Cvaleur>[_nvct];
  for (j=0;j<_nvct;j++)
    _seq[j].construit_a_vide(_nval);
  
  _fin=&_seq[_nvct-1];
  
  _l_desc.construit_a_vide(_nval);
  for (j=0;j<_nval;j++)
    _l_desc[j]=s[j];
}


void Cmatrice::operator=(const Fmatrice& M)
{
  int i;
  if ((_nval!=M._nval) || (_nvct!=M._nvct)){
    nettoie();
    if (_seq)
      delete[] _seq;
    _nval=M._nval;
    _nvct=M._nvct;
    if (_nvct){
      _seq = new vector<Cvaleur>[_nvct];
      for (i=0;i<_nvct;i++)
        _seq[i].construit_a_vide(_nval);

      _fin=&_seq[_nvct-1];
    }
  }
  if (!_nvct){
    _seq=0;
    _fin=0;
    return;
  }

  _l_desc.recup_prem(M._l_desc,_nval);

  for (i=0;i<_nvct;i++)
    _seq[i].recup_prem(M._seq[i],_nval);
}

bool Cmatrice::recup_mat(const string& ns)
{
  int i,j,k;
  int deb;

  if (ns[0]=='\0')
    return 0;

  ifstream flot(ns.c_str());
  if (!flot){
    cerr << "Pb opening " << ns << endl;
    return 0;
  }

  nettoie();

  //////////// rcup de la longueur des squences de param 

  string str;

  do {
    getline(flot,str);
  } while (!flot.eof() && !isdigit(str[0]) && (str[0]!='('));

  if (flot.eof()){
      cerr << "Bad header: ";
      cerr << ns << endl;
      flot.close();
      return 0;
  }
    
  deb=(str[0]=='(')?atoi((str.substr(1)).c_str()):atoi(str.c_str());

  if (deb<=0){
    cerr << "Bad start in file ";
    cerr << ns << endl;
    flot.close();
    return 0;
  }

  i=str.find_first_of(" \t");
  i=str.find_first_not_of(" \t",i);

  if (i!=string::npos){
    k=atoi((str.substr(i)).c_str());
    if ((_nvct=1+k-deb)<=0){
      cerr << "Bad boundaries in file ";
      cerr << ns << endl;
      flot.close();
      return 0;
    }
  }
  else{
    _nvct=deb;
    deb=1;
  }

  ///////////// teste l'tat des info du fichier ////////////////
  ///////////////// rcuperation des donnes ////////////////////

  do{
    getline(flot,str);
  } while (!flot.eof() && 
	   (str.find_first_not_of(" \t")==string::npos));

  if (flot.eof()){
    cerr << "Problem in file "<< ns << endl;
    flot.close();
    return 0;
  }

  _nval=0;
  i=0;
  bool a=1;
  while (i<str.length()){
    if (!isspace(str[i++])){
      if (a)
        _nval++;
      a=0;
    }
    else 
      a=1;
  }

  if (_nval<=0){
    cerr << "Characters missing " << endl;
    flot.close();
    return 0;
  }
    
  _l_desc.construit_a_vide(_nval);
  i=0;k=0;
  while (i<str.length()){
    if (!isspace(str[i])){
      if (k==_nval){
	cerr << "Error: too many characters " << endl;
        _nval=0;
	_l_desc.construit_a_vide(0);
	flot.close();
	return 0;
      }
      if (isalpha(str[i]))
        _l_desc[k++]=str[i];
      else if (str[i]=='#'){
        j=atoi((str.substr(i+1)).c_str());
        if ((j<0) || (j>=256)){
          cerr << "Error: out of range character " << j << endl;
          _nval=0;
          _l_desc.construit_a_vide(0);
          flot.close();
          return 0;
        }          
        _l_desc[k++]=(char)j;
        i=str.find_first_of(" \t\n",i);
      }
      else {
        cerr << "Problem in characters declaration " << endl;
        _nval=0;
	_l_desc.construit_a_vide(0);
	flot.close();
	return 0;
      }
    }
    i++;    
  }
  
  _seq = new Cmatrice::vector<Cvaleur>[_nvct];
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);

  _fin=_seq+(_nvct-1);  

  i=deb-1;
  while ((!flot.eof()) && (i>0)){
    getline(flot,str);
    if (str.find_first_not_of(" \t")!=string::npos)
      i--;
  }
   
  if (flot.eof()){
    cerr << "File " << ns << " returns empty tabular." << endl;
    nettoie();
    flot.close();
    return 0;
  }

  i=0;
  while (!flot.eof() && (i<_nvct)){
    getline(flot,str);
    if (str.find_first_not_of(" \t")!=string::npos){
      k=0;
      j=0;
      while (k<_nval){
	_seq[i][k++].g_val(atof((str.substr(j)).c_str()));
	j=str.find_first_of(" \t",j);
	if ((
	     (j=str.find_first_not_of(" \t",j))==string::npos)
	    && (k<_nval)){
	  cerr << "Error: line " << i+deb << " is too short ";
	  cerr << "in file " << ns << endl;
	  nettoie();
	  flot.close();
	}
      }
      i++;
    }
  }
  
  if (i<_nvct){
    cout << "Too short Matrice..." << endl;
    _nvct=i;
  }
  flot.close();
  return 1;
}

void Cmatrice::FB(const Fsequence& G, const Clexique& L)
{
  if (G.deb()==0)
    return;
  
  int ta=G.taille();
  int i;
  int ndesc;
  
  Tliste<int> ldesc;
  L.ls_num(ldesc);
  
  Tliste_iter<int> iti(ldesc);
  int* pi;
  
  if (!(ndesc=ldesc.taille()))
    return;

  // allocation de la Cmatrice
  
  _nval=ndesc;
  _nvct=ta;

  if (_seq)
    delete[] _seq;
  _seq = new vector<Cvaleur>[_nvct];
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);

  _fin=_seq+_nvct-1;
  _l_desc.construit_a_vide(_nval);
  
  i=0;
  iti.init();
  while (pi=iti())
    _l_desc[i++]=int(*pi);
  
  // remplissage par FB

  RM_lex<Cposition> RL(L);

  const Cdescripteur *di, *dd, *df;
  di=G.deb();
  dd=df=G.fin();

  ///////////////////////////
  // Remplit le tableau du backward

  int lg=G.taille();
  int l=lg;
  
  Tliste<double>* lval=new Tliste<double>[lg+1];
  Tliste_iter<double> itd1, itd2;
  
  RL.init();
  RL.aj_val(lval[l--]);
  
  do {
    RL.backward(di,dd,df);
    RL.aj_val(lval[l--]);
  } while(--dd>di);

  
  /////////////////////////////
  // Realise le forward

  dd=di;
  l=0;
  
  RL.init(di,dd,df);
  RL.aj_val(lval[l]); 
  double m,x;
  
  while (dd++<df){
    itd1.colle(lval[l]);
    itd2.colle(lval[l+1]);
    m=*(itd1++)+*(itd2++);
    g_val(m,0,l);
    for (i=1;i<_nval;i++){
      g_val(x=*(itd1++)+*(itd2++),i,l);
      if (m<x)
	m=x;
    }
    x=0;
    for (i=0;i<_nval;i++)
      x+=exp(_seq[l][i].val()-m);
    x=log(x)+m;
    for (i=0;i<_nval;i++)
      _seq[l][i]-=x;
    
    RL.forward(di,dd,df);
    RL.maj_val(lval[++l]);
  }

  itd1.colle(lval[l]);
  itd2.colle(lval[l+1]);
  m=*(itd1++)+*(itd2++);
  g_val(m,0,l);
  for (i=1;i<_nval;i++){
    g_val(x=*(itd1++)+*(itd2++),i,l);
    if (m<x)
      m=x;
  }
  x=0;
  for (i=0;i<_nval;i++)
    x+=exp(_seq[l][i].val()-m);
  x=log(x)+m;
  for (i=0;i<_nval;i++)
    _seq[l][i]-=x;

  delete[] lval;
}

void Cmatrice::forward(const Fsequence& G, const Clexique& L)
{
  if (G.deb()==0)
    return;
  
  int ta=G.taille();
  int i;
  int ndesc;
  
  Tliste<int> ldesc;
  L.ls_num(ldesc);
  
  Tliste_iter<int> iti(ldesc);
  int* pi;
  
  if (!(ndesc=ldesc.taille()))
    return;

  // allocation de la Cmatrice
  
  _nval=ndesc;
  _nvct=ta;

  if (_seq)
    delete[] _seq;
  _seq = new vector<Cvaleur>[_nvct];
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);
  
  _fin=_seq+_nvct-1;
  _l_desc.construit_a_vide(_nval);
  
  i=0;
  iti.init();
  while (pi=iti())
    _l_desc[i++]=int(*pi);
  
  // remplissage par forward

  RM_lex<Cposition> RL(L);

  const Cdescripteur *di, *dd, *df;
  dd=di=G.deb();
  df=G.fin();
  int l=0;
  
  Tliste<double> lval;
  Tliste_iter<double> itd(lval);
  
  RL.init(di,dd,df);
  RL.aj_val(lval);
  itd.init();
  for (i=0;i<_nval;i++)
    g_val(*(itd++),i,0);
  
  while (dd++<df){
    l++;
    RL.forward(di,dd,df);
    RL.maj_val(lval);
    
    itd.init();
    for (i=0;i<_nval;i++){
      g_val(*(itd++),i,l);
    }
  }
}

void Cmatrice::backward(const Fsequence& G, const Clexique& L)
{
  if (G.deb()==0)
    return;
  
  int ta=G.taille();
  int i;
  int ndesc;
  
  Tliste<int> ldesc;
  L.ls_num(ldesc);
  
  Tliste_iter<int> iti(ldesc);
  int* pi;
  
  if (!(ndesc=ldesc.taille()))
    return;

  // allocation de la Cmatrice
  
  _nval=ndesc;
  _nvct=ta;

  if (_seq)
    delete[] _seq;
  _seq = new vector<Cvaleur>[_nvct];
  
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);
  _fin=_seq+_nvct-1;
  _l_desc.construit_a_vide(_nval);
  
  i=0;
  iti.init();
  while (pi=iti())
    _l_desc[i++]=int(*pi);
  
  ////////////////////////////////////
  // remplissage par backward

  RM_lex<Cposition> RL(L);

  const Cdescripteur *di, *dd, *df;
  di=G.deb();
  dd=df=G.fin();

  int l=G.taille()-1;
  
  Tliste<double> lval;
  Tliste_iter<double> itd(lval);
  
  RL.init();
  RL.aj_val(lval);
  
  itd.init();
  for (i=0;i<_nval;i++)
    g_val(*(itd++),i,l);
  
  while (dd>di){
    l--;
    RL.backward(di,dd--,df);
    RL.maj_val(lval);

    itd.init();
    for (i=0;i<_nval;i++)
      g_val(*(itd++),i,l);
  }
}


void Cmatrice::compresse(const Fsequence& G, const int f, char op)
{
  if (f<=0)
    return;

  int ta=G.taille();
  int i;

  Tliste<Cdescripteur> ld;
  Tliste_iter<Cdescripteur> ild(ld);
  Cdescripteur* pd;

  const Cdescripteur *dd, *df, *di;

  // allocation de la Cmatrice et du lexique

  for (i=0;i<ta;i++)
    ld.aj_fin_diff(G[i]);

  nettoie();

  _nval=ld.nb_ind();
  _nvct=ta/f;
  
  _seq = new vector<Cvaleur>[_nvct];
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);

  _fin=&_seq[_nvct-1];
  _l_desc.construit_a_vide(_nval);
  
  i=0;
  ild.init();
  while (pd=ild())
    _l_desc[i++]=*pd;

  di=dd=G.deb();
  df=G.fin();
  i=0;
  while (i<_nvct) {
    di+=f;
    do{
      ajoute_en(i,*dd);
    } while (++dd<di);
    if (op=='m')
      divise_en_par(i,f);
    i++;
  }
}

void Cmatrice::compresse(const Fmatrice& M, const int f, char op)
{
  if (f<=0)
    return;

  int ta=M.taille();
  int i,j;

  Tliste<Cdescripteur> ld;
  Tliste_iter<Cdescripteur> ild(ld);
  Cdescripteur* pd;

  mat_iter mi, mi2;

  // allocation de la Cmatrice et du lexique

  j=M.nval();
  for (i=0;i<j;i++)
    ld.aj_fin_diff(M.dsc(i));

  nettoie();
  
  _nval=ld.nb_ind();
  _nvct=ta/f;
  
  _seq = new vector<Cvaleur>[_nvct];
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);
  _fin=&_seq[_nvct-1];
  _l_desc.construit_a_vide(_nval);
  
  i=0;
  ild.init();
  while (pd=ild())
    _l_desc[i++]=*pd;

  //  remplissage de la Cmatrice

  mi.colle(M);
  mi2.colle(M);
  mi2+=f-1;
  i=0;
  while (i<_nvct){
    ajoute_en(i,mi);
    while (mi!=mi2){
      mi++;
      ajoute_en(i,mi);
    }
    if (op=='m')
      divise_en_par(i,f);
    mi2+=f;
    mi++;
    i++;
  }
}

void Cmatrice::variation(const Fsequence& G)
{
  int ta=G.taille();
  int i;

  Tliste<Cdescripteur> ld;
  Tliste_iter<Cdescripteur> ild(ld);
  Cdescripteur* pd;
  
  const Cdescripteur *dd, *df;

  // allocation de la Cmatrice et du lexique

  for (i=0;i<ta;i++)
    ld.aj_fin_diff(G[i]);
  
  _nval=ld.nb_ind();
  _nvct=ta;

  if (_seq)
    delete[] _seq;
  _seq = new vector<Cvaleur>[_nvct];
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);

  _fin=_seq+_nvct-1;
  _l_desc.construit_a_vide(_nval);
  
  i=0;
  ild.init();
  while (pd=ild())
    _l_desc[i++]=*pd;

  //  remplissage de la Cmatrice

  dd=G.deb();
  df=G.fin();
  ajoute_en(0,*dd);
  i=1;
  do {
    ote_en(i,*dd);
    dd++;
    ajoute_en(i,*dd);
    i++;
  } while (dd<df);
}

void Cmatrice::variation(const Fmatrice& M)
{
  int ta=M.taille();
  int i,j;

  Tliste<Cdescripteur> ld;
  Tliste_iter<Cdescripteur> ild(ld);
  Cdescripteur* pd;
  
  mat_iter mi;

  // allocation de la Cmatrice et du lexique

  j=M.nval();
  for (i=0;i<j;i++)
    ld.aj_fin_diff(M.dsc(i));

  _nval=ld.nb_ind();
  _nvct=ta;

  if (_seq)
    delete[] _seq;
  _seq = new vector<Cvaleur>[_nvct];
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);

  _fin=_seq+_nvct-1;
  _l_desc.construit_a_vide(_nval);
  
  i=0;
  ild.init();
  while (pd=ild())
    _l_desc[i++]=*pd;

  //  remplissage de la Cmatrice

  mi.colle(M);
  i=1;
  ajoute_en(0,mi);
  do{
    ote_en(i,mi);
    mi++;
    ajoute_en(i,mi);
    i++;
  } while (i<_nvct);
}

///////////////////////////////////////////////////////////////////

void Cmatrice::integre(const Fsequence& G)
{
  int ta=G.taille();
  int i;

  Tliste<Cdescripteur> ld;
  Tliste_iter<Cdescripteur> ild(ld);
  Cdescripteur* pd;
  
  const Cdescripteur *dd, *df;

  // allocation de la Cmatrice et du lexique

  for (i=0;i<ta;i++)
    ld.aj_fin_diff(G[i]);
  
  _nval=ld.nb_ind();
  _nvct=ta;

  if (_seq)
    delete[] _seq;
  _seq = new vector<Cvaleur>[_nvct];
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);

  _fin=_seq+_nvct-1;
  _l_desc.construit_a_vide(_nval);
  
  i=0;
  ild.init();
  while (pd=ild())
    _l_desc[i++]=*pd;

  mat_iter mi;

  //  remplissage de la Cmatrice

  dd=G.deb();
  df=G.fin();
  mi.colle(*this);

  ajoute_en(0,*dd);
  dd++;
  i=1;
  while (dd<=df){
    ajoute_en(i,mi);
    mi++;
    ajoute_en(i,*dd);
    i++;
    dd++;
  } 
}

void Cmatrice::integre(const Fmatrice& M)
{
  int ta=M.taille();
  int i,j;

  Tliste<Cdescripteur> ld;
  Tliste_iter<Cdescripteur> ild(ld);
  Cdescripteur* pd;
  
  mat_iter mi;

  // allocation de la Cmatrice et du lexique

  j=M.nval();
  for (i=0;i<j;i++)
    ld.aj_fin_diff(M.dsc(i));

  _nval=ld.nb_ind();
  _nvct=ta;

  if (_seq)
    delete[] _seq;
  _seq = new vector<Cvaleur>[_nvct];
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);

  _fin=_seq+_nvct-1;
  _l_desc.construit_a_vide(_nval);
  
  i=0;
  ild.init();
  while (pd=ild())
    _l_desc[i++]=*pd;

  //  remplissage de la Cmatrice

  mat_iter mi2;
  mi2.colle(*this);

  mi.colle(M);

  ajoute_en(0,mi);
  i=1;
  while(mi()){
    ajoute_en(i,mi2);
    mi2++;
    ajoute_en(i,mi);
    i++;
  } 
}

void Cmatrice::traduit(const Fsequence& D, const Clexique& L)
{
  int ta=D.taille();
  int i;

  Tliste<Cdesc_comp> ld;
  Tliste_iter<Cdesc_comp> ild(ld);
  Cdesc_comp* pd;

  const Cdescripteur *dds, *dd, *df;

  L.liste_desc(ld);

  nettoie();
 
  // Vrification des bons numros
  
  if ((_nval=ld.nb_ind())==0){
    cerr << "Empty Lexique!" << endl;
    return;
  }

  // Allocation et construction
  
  _nvct=ta;
  if ((_seq = new vector<Cvaleur>[_nvct])==0){
    cerr << "Allocation error in Cmatrice::traduit" << endl;
    return;
  }
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);

  _fin=_seq+_nvct-1;
  _l_desc.construit_a_vide(_nval);

  ild.init();
  i=0;
  while (pd=ild())
    _l_desc[i++].g_dsc((char)(pd->num()));

  //// remplissage de la Cmatrice
  
  vector<Cvaleur>* pv=_seq;
  
  dds=dd=D.deb();
  df=D.fin();
  do{
    ild.init();
    i=0;
    while (pd=ild())
      (*pv)[i++]=pd->multiplie(dds,dd,df);
    pv++;
  } while (dd++<df);
  
}


void Cmatrice::traduit(const Fmatrice& D, const Clexique& L)
{
  int ta=D.taille();
  int i;

  Tliste<Cdesc_comp> ld;
  Tliste_iter<Cdesc_comp> ild(ld);
  Cdesc_comp* pd;

  mat_iter mi;

  L.liste_desc(ld);

  nettoie();
 
  // Vrification des bons numros
  
  if ((_nval=ld.nb_ind())==0){
    cerr << "Empty Lexique!" << endl;
    return;
  }

  // Allocation et construction
  
  _nvct=ta;
  if ((_seq = new vector<Cvaleur>[_nvct])==0){
    cerr << "Allocation error in Cmatrice::traduit" << endl;
    return;
  }
  for (i=0;i<_nvct;i++)
    _seq[i].construit_a_vide(_nval);

  _fin=_seq+_nvct-1;
  _l_desc.construit_a_vide(_nval);

  ild.init();
  i=0;
  while (pd=ild())
    _l_desc[i++].g_dsc((char)(pd->num()));

  //// remplissage de la Cmatrice

  vector<Cvaleur>* pv=_seq;

  mi.colle(D);
  do{
    ild.init();
    i=0;
    while (pd=ild())
      (*pv)[i++]=pd->multiplie(mi);
    pv++;
  } while (mi());
  
}


////////////////////////////////
// modifications

void Cmatrice::ajoute_en(int i, const mat_iter& m)
{
  int a,j;
  if ((i>=0) && (i<_nvct))  
    for (j=0;j<m.mat()->nval();j++)
      if ((a=ligne(m.mat()->dsc(j)))>=0)
	_seq[i][a]+=m(j);
}

void Cmatrice::ote_en(int i, const mat_iter& m)
{
  int a,j;
  if ((i>=0) && (i<_nvct))  
    for (j=0;j<m.mat()->nval();j++)
      if ((a=ligne(m.mat()->dsc(j)))>=0)
	_seq[i][a]-=m(j);
}

