// Copyright 2004 Laurent GUEGUEN

// This file is part of Sarment.

// Sarment is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.

// Sarment is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with Sarment; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#if !defined _MATRIX_H_
#define _MATRIX_H_

#include <string>
#include <sstream>

#include "Fmatrice.h"

using std::string;

class mat_iter;
class Cpartition;
class Clexique;
class Fsequence;
class Cvaleur;

class Cmatrice : public Fmatrice
{
public:

  Cmatrice(const Fmatrice&);
  Cmatrice(void*) {};
  
  inline Cmatrice();
  inline ~Cmatrice();
  inline void nettoie();

  // genere une matrice de longueur int_1, dont les descripteurs
  // sont les caracteres de str_1
  void genere(const int, const string&);
  
  //remplie par FB
  void FB(const Fsequence&, const Clexique&);
  
  //remplie par forward
  void forward(const Fsequence&, const Clexique&);
  
  //remplie par backward  
  void backward(const Fsequence&, const Clexique&);
  
  //////////////////////////////////
  //  modifications
  
  bool recup_mat(const string&);

  
  ///////////
  // sur Clexique_1 SANS MOTIFS, construit une Cmatrice sur un alphabet
  // dont chaque lettre correspond a un predicteur de Clexique_1 (au
  // plus 52), et en chaque position a la valeur de la prediction du
  // predicteur de la ligne en la position de Csujet_1 
  void traduit(const Fmatrice&, const Clexique&);
  void traduit(const Fsequence&, const Clexique&);
  
  ///////
  // par la compression de Csujet_1 en paquets
  //   si char_1=='m', fait la moyenne
  //   sinon fait la somme
  // paquets de taille int_1
  void compresse(const Fsequence&, const int, char);
  void compresse(const Fmatrice&, const int, char);


  void variation(const Fsequence&);
  void variation(const Fmatrice&);
  
  void integre(const Fsequence&);
  void integre(const Fmatrice&);

  // ajoute la donnee $2 en $1
  inline void ajoute_en(int, const Cdescripteur&);
  void ajoute_en(int, const mat_iter&);

  // soustrait la donnee $2 en $1
  inline void ote_en(int, const Cdescripteur&);
  void ote_en(int, const mat_iter&);

  // divise en $1 par la valeur $2
  inline void divise_en_par(int, double);

  //copie
  void operator=(const Fmatrice&);

  ///////////////////////////////////
  //  donnees

  inline Fmatrice::vector<Cvaleur>& operator[](const int) const;

  ////////////////////////////
  //  ordres 

  friend class mat_iter;

  /////////////////////////
  /////// sorties      ////

  void ecrit_evalue() const;
};


#include "matrice.inl"
#endif
